// file:        mbus-interface.C
// created by:  David Lauzon, University of Toronto
// date:        March 29, 1994.
// version:     0.0
// last update: March 29, 1994.



// This module is an interface to the Message Bus. The Message Bus provides
// a software "bus" to allow multiple client programs to communicate
// simultaneously, using selective broadcast messages.  Clients communicate
// by indicating interest in various classes of messages, and then sending 
// messages with (possibly different) classes to the MBus, which then 
// forwards the message to all clients that are interested in it. Message
// delivery is guaranteed to be atomic in the sense that a message is received
// completely or not at all, and that the entirety of a message wil be 
// delivered to a client before any part of any subsequent message.
//
// Messages are sent in an S-expression format, and come in two forms,
// transmissions and commands. Transmissions are messages that are intended 
// to be received by other clients, while commands are messsages to the MBus 
// itself.
//
// Every transmission must have the form:
//
//        transmission:  (tag domains data)
//        where tag          -- is not really of any use in the current 
//                             implementation.
//              domains      -- is a list of domainSpecs
//              data         -- 1 or more S-expressions
//              domainSpecs  -- field[.field[.field]]
//              field        -- context | tool class | user id
//                NOTE for multiple fields, the ordering is important: 
//                     context before tool class before user id.
//              s-expression -- string
//
// Strings and fields are sequences of characters with '"' and '\' escaped
// by '\'. The '\' character is removed when the transmission is read by
// the MBus for purposes of comparison, but replaced when the transmission is
// sent to a client.
//
// MB_connect(host, port) --> Boolean
//    Create a connection to a message MBus. host is a null terminating string
//    specifying the name of the host where the MBus is running, and port is 
//    the port number the MBus is listening at.
//
// MB_close() --> Boolean
//    Close the connection to the MBus. When this command is executed, the 
//    client is removed from the internal MBus tables, and the socket 
//    connection is closed.
//
// MB_Clear() --> Boolean
//    Clear all expressed interest. After a clear, a client will not receive
//    any messages until a MB_accept is sent.
//
// MB_id(userId) --> Boolean
//    Set the identifier for the client.
//
// MB_accept(domainList) --> Boolean
//    Express interest in a class of messages. The domainList is a list of
//    domain expressions that describe interesting domains. Each expression
//    is similar in form to a domain specification for a message, but a
//    leading and/or '.' is allowed, and fields may be omitted. An ommitted
//    field is matched by any field in a message. A leading '.' indicates 
//    that any number (including 0) of fields to the left are acceptable, 
//    while a trailing '.' indicates that any number of fieldsto the right
//    are acceptable.
//
// MB_sendMessage(message, domain)
//    Write a buffer to an MBus connection.
//
// MB-waitForMessage() --> S-expression
//    Read the next message from the MBus connection. The socket connection
//    is set to non-blocking by MBConnect, so that if there is not a complete
//    message waiting, the function returns immediately. 
//    Returns 
//       NULL if there is not a complete message waiting on the connection.
//       A pointer to an S-expression if a message was ready.
//
// MB_getDomainString(context, userId, toolClass) --> DomainString
//    Translate the the tml domain representation to MBus domain 
//    representation.
//
// MB_getDomainSpecString(context, userId, toolClass) --> DomainSpecString
//    Translate the the tml domain representation to MBus' representation of
//    domain specification.


#include "os.H"

void MB_connect(String* host, int port);
void MB_close();
void MB_clear();
void MB_id(String* userId);
void MB_accept(String* domains);
void MB_sendMessage(String* c, String* d);
t_sexp  MB_waitForMessage();
t_sexp  MB_getNewMessage();
String* MB_makeDomainString(String* context,String* userId, String* toolClass);
String* MB_makeDomainSpecString(String* context,String* userId, String* toolClass);
