//
// File:    graphnode.c
// Author:  Brian Corrie
// Date:    January 17, 1991
// Project: Rigi V4.0
//

// $Author: jmartin $
// $Date: 2000-01-12 01:14:02 $
// $Source: /root/cvs/Rigi/rigiedit/cl_graphnode.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: cl_graphnode.C,v 1.53 2000-01-12 01:14:02 jmartin Exp $";

//------------------------------------------------------------------------

#ifndef STREAMH
#include <stream.h>
#endif

// Basic class definitions

#ifndef GRAPHARCLIST_H
#include "cl_grapharclist.h"
#endif
#ifndef GRAPHARCSET_H
#include "cl_grapharcset.h"
#endif
#ifndef HASHNODELIST_H
#include "cl_hashnodelist.h"
#endif

#ifndef GRAPHNODE_H
#include "cl_graphnode.h"	// Graph node class definition
#endif
#ifndef GRAPHARC_H
#include "cl_grapharc.h"	// Graph arc class definition
#endif
#ifndef ARCFLAGS_H
#include "cl_arcflags.h"        // ArcFlags class definition
#endif
#ifndef ATTRIBUTE_H
#include "cl_attribute.h"	// For the attribute set of this node
#endif
#ifndef GRAPHERR_H
#include "grapherr.h"           // Want to include low level stdio error funcs
#endif

// Global declarations for the editor state and utility/interface functions

#ifndef GLOBAL_H
#include "global.h"		// for editor and ANY_ARC
#endif
#ifndef RIGISTATE_H
#include "cl_rigistate.h"       // Global instance of editor's state
#endif

#ifndef GREP_H
#include "grep.h"		// For Unix like grep functionality
#endif

// Define some of the constants that are declared in cl_graphdef.h

const Coordinate IllegalPos     = -999;
const Point      IllegalPoint   = { IllegalPos,IllegalPos,IllegalPos };
const NDC_Coord  NDC_IllegalPos = -999.0F;

static int nnodes=0;


// Constructor for GraphNode performs all the initialization

GraphNode::GraphNode( NodeLabel label ) {
    nnodes++;
    nLa = NULL;
    nId = editor->NextObjId();
    info = 0;
    pos.x = pos.y = 0.0F;
    active = NULL;

    SetLabel( label );

    attributes = NULL;

    srcarcs = new ArcSet;
    dstarcs = new ArcSet;
}

// Destructor for Nodes. Before a node is deleted, it is necessary to clean
// up the lists of arcs. There should be no arcs in any of the arc list for
// the node. It should also be checked to make sure that no RigiWindows in
// the system contain a reference to the node. GraphNode should not have to know
// about these details. It should be up to the application that uses a GraphNode
// to clean this up.

int GetNoGraphNodes() { return nnodes; }

GraphNode::~GraphNode() {
    nnodes--;
    // Free up the memory associated with the node label and attributes

    if ( nLa ) delete nLa;
    if (attributes) delete attributes;

    // MJW  The deleter of this node is responsible for making sure
    // the arcs contained in the node's lists are all cleaned up first.
    // This should perhaps be done in WindowDeleteNode().

    delete srcarcs;
    delete dstarcs;
}

void GraphNode::SetId(NodeId id) {
    nId = id;
    if (id > editor->GetObjId()) editor->SetObjId(id);
}

void GraphNode::SetType( NodeType t) {
    if (t<0 || t>0x7fff) mayday("bad nodetype in GraphNode::SetType()");
    info &= 0xffff; // keep the lower 16 bits, clear the rest
    info |= t<<16;  // set the higher 16 bits with the value
}

void GraphNode::SetLevel( int i ) {
    if (i<0 || i>0x7fff) mayday("bad level in GraphNode::SetLevel(int)");
    info &= ~(0x7fff)<<1;
    info |= i<<1;
}

void GraphNode::SetLoaded( int i ) {
    if (i) info |= 1; else info = (info>>1)<<1;
}


// Make a local copy of the label passed in, deleting old one if necessary

void GraphNode::SetLabel( NodeLabel label ) {
    if (nLa) delete nLa;
    if (!label)
        nLa = NULL;
    else {
        nLa = new char[strlen(label)+1];
        if (!nLa) { mayday("No memory for node label????"); exit(1); }
        strcpy( nLa,label );
    }
}

// Set level to 1 more than the greatest level of its parents

void GraphNode::SetLevel() {
    int maxlevel = 0;
    AvlPos *P;
    GraphArc* ap = FirstAdj( P, inLevArcs );

    while ( ap ) {
        GraphNode* src = ap->GetSrc();
        if (src->GetLevel()+1 > maxlevel) {
            maxlevel = src->GetLevel()+1;
        }
        ap = NextAdj( P, inLevArcs );
    }
    SetLevel(maxlevel);
}

// Set the Normalized position of the GraphNode. Check for boundary errors.
// The values should be between 0 and 1.

void GraphNode::SetNDCPosition( NDC_Point2d p ) {
    pos = p;
}

// Add to the list passed in all the predecessor nodes of this node

void GraphNode::GetPredecessor( HashNodeList& nlist ) {
    AvlPos *P;
    GraphArc* ap = FirstAdj( P, inArcs );
    while( ap != NULL ) {
        nlist.Add( *(ap->GetSrc()) );
        ap = NextAdj( P, inArcs );
    }
}

// Add to the list passed in all the successor nodes of this node

void GraphNode::GetSuccessor( HashNodeList& nlist ) {
    AvlPos *P;
    GraphArc* ap = FirstAdj( P, outArcs );
    while( ap != NULL ) {
        nlist.Add( *(ap->GetSrc()) );
        ap = NextAdj( P, outArcs );
    }
}

// Add to the list passed in all the predecessor nodes of this node, if and
// only if the arc that joins them is of the given type. Note it
// is an error to get the predecessors and to specify inArcs of any type.

void GraphNode::GetPredecessor( HashNodeList&nlist,AdjList list,ArcType type ) {
    if ( list != inArcs && list != inLevArcs )
        mayday("GraphNode::GetPredecessor - Invalid arc type");

    AvlPos *P;
    GraphArc* ap = FirstAdj( P, list );
    while ( ap != NULL ) {
        if ( (list!=inArcs&&list!=outArcs)
            || ANY_ARC==type || ap->GetType()==type )
            nlist.Add( *(ap->GetSrc()) );
        ap = NextAdj( P, list );
    }
}

// Add to the list passed in all the successor nodes of this node, if and
// only if the arc that joins them is of the given type. Note it
// is an error to get the successors and to specify outArcs of any type.

void GraphNode::GetSuccessor( HashNodeList& nlist, AdjList list, ArcType type ) {
    if ( list != outArcs && list != outLevArcs )
        mayday("GraphNode::GetSuccessor - Invalid arc type");

    AvlPos *P;
    GraphArc* ap = FirstAdj( P, list );
    while( ap != NULL ) {
        if ( (list!=inArcs&&list!=outArcs)
            || ANY_ARC==type || ap->GetType()==type )
            nlist.Add( *(ap->GetDst()) );
        ap = NextAdj( P, list );
    }
}


// comparison function for AVL sets of arcs

static GraphArc *dummyarc = NULL;

static int cmparc( GraphArc *a, GraphArc *b ) {
    if (a->GetType() > b->GetType()) return  1;
    if (a->GetType() < b->GetType()) return -1;

    if (a->GetDst() > b->GetDst()) return  1;
    if (a->GetDst() < b->GetDst()) return -1;

    if (a->GetSrc() > b->GetSrc()) return  1;
    if (a->GetSrc() < b->GetSrc()) return -1;

    return 0;
}

// Insert an arc into the specified list.

void GraphNode::InsertArc( GraphArc* ap ) {
    if (ap->GetSrc()==this) srcarcs->Add(ap,cmparc);
    if (ap->GetDst()==this) dstarcs->Add(ap,cmparc);
}

// Remove an arc from the specified list

GraphArc* GraphNode::RemoveArc( GraphArc* ap ) {
    GraphArc *a;
    if (ap->GetSrc()==this) a = srcarcs->Remove(ap,cmparc);
    if (ap->GetDst()==this) a = dstarcs->Remove(ap,cmparc);

    if (!a) {
        mayday("can't remove GraphArc %d from set",ap);
        mayday("(arc src %s, dst %s)",
            ap->GetSrc()->GetLabel(),ap->GetDst()->GetLabel());
    }
    return a;
}

GraphArc* GraphNode::FindArc( GraphNode *n, AdjList type ) {
    if (!n) return NULL;

    AvlPos *P;
    for ( GraphArc *a=FirstAdj(P,type); a; a=NextAdj(P,type) )
        if ( a->GetOtherNode(this)==n ) return a;
    return NULL;
}

GraphArc* GraphNode::FindArc( GraphNode*, AdjList, ArcType ) {
    mayday("FindArc unimplemented!");
    return NULL;
}

GraphArc* GraphNode::FirstArc() {
    GraphArc *a = srcarcs->First();
    if (a == NULL) a = dstarcs->First();
    return a;
}

GraphArc* GraphNode::FirstArc( const AdjList adj ) {
    GraphArc *a = NULL;
    AvlPos *P; // memory leak!

    a = FirstAdj( P, adj );
    return a;
}

GraphArc* GraphNode::FirstAdj( AvlPos* &P, const AdjList adj ) {
    int lt = editor->LevelArc();

    ArcDirection dir;
    if (adj==inLevArcs || adj==inArcs) dir = ARC_IN;
    else if (adj==outLevArcs || adj==outArcs) dir = ARC_OUT;
    else mayday("problem in FirstAdj");

    if ( adj==inLevArcs || adj==outLevArcs )
        return FirstArc( P,lt,dir );


    ArcSet *arcset = (dir==ARC_IN) ? dstarcs : srcarcs;

    if (dummyarc == NULL) dummyarc =  new GraphArc(NULL, NULL, 0);
    dummyarc->SetSrc( (dir==ARC_IN) ? NIL : this );
    dummyarc->SetDst( (dir==ARC_IN) ? this : NIL );
    dummyarc->SetType(0);

    GraphArc *a = arcset->First( P, dummyarc, cmparc );
    if (a && a->GetType()==lt) {
	arcset->Delete(P);
        dummyarc->SetType(lt+1);
        a = arcset->First( P, dummyarc, cmparc );
    }

    // MJW  Before we return, make the dummyarc safe
    dummyarc->SetSrc(NULL);
    dummyarc->SetDst(NULL);

    return a;
}

GraphArc* GraphNode::NextAdj( AvlPos* &P, const AdjList adj ) {
    int lt = editor->LevelArc();

    //ArcDirection dir = (adj==inLevArcs || adj==inArcs) ? ARC_IN : ARC_OUT;
    ArcDirection dir;
    if (adj==inLevArcs || adj==inArcs) dir = ARC_IN;
    else if (adj==outLevArcs || adj==outArcs) dir = ARC_OUT;
    else mayday("problem in NextAdj");

    if ( adj==inLevArcs || adj==outLevArcs )
        return NextArc( P,lt,dir );

    ArcSet *arcset = (dir==ARC_IN) ? dstarcs : srcarcs;

    GraphArc *a = arcset->Next(P);
    if (a && a->GetType()==lt ) {
        arcset->Delete(P);
        if (dummyarc == NULL) dummyarc =  new GraphArc(NULL, NULL, 0);
        dummyarc->SetSrc( (dir==ARC_IN) ? NIL : this );
        dummyarc->SetDst( (dir==ARC_IN) ? this : NIL );
        dummyarc->SetType( lt+1 );

        a = arcset->First( P, dummyarc, cmparc );

        // MJW  Before we return, make the dummyarc safe
        dummyarc->SetSrc(NULL);
        dummyarc->SetDst(NULL);
    }
    return a;
}

GraphArc* GraphNode::FirstArc(AvlPos *&P, ArcType type, ArcDirection &dir) {
    ArcSet *arcset;
    if (dummyarc == NULL) dummyarc =  new GraphArc(NULL, NULL, 0);
    if (dir == ARC_IN || dir == ARC_INOUT_IN) {
        arcset = dstarcs;
        dummyarc->SetSrc(NULL);
        dummyarc->SetDst(this);
    } else {
        arcset = srcarcs;
        dummyarc->SetSrc(this);
        dummyarc->SetDst(NULL);
    }
    dummyarc->SetType((type == ANY_ARC)? 0 : type );

    GraphArc *a = arcset->First( P, dummyarc, cmparc );

    if (dir == ARC_INOUT) {
        if (a == NULL || (type != ANY_ARC && a->GetType() != type)) {
            if (a != NULL)
	      srcarcs->Delete(P);
            // nothing in srcarcs, look in dstarcs
            dir = ARC_INOUT_IN;
            dummyarc->SetSrc( NULL);
            dummyarc->SetDst( this);
            a = dstarcs->First(P, dummyarc, cmparc);
        }
    } else if (dir == ARC_INOUT_IN) {
        // check for duplication of self-arcs.
        // note that although there should be only one,
        // this code is able to handle many
        while (a != NULL && a->GetSrc() == this) {
            a = dstarcs->Next(P);
        }
    }

    // Before we return, make the dummyarc safe
    dummyarc->SetSrc(NULL);
    dummyarc->SetDst(NULL);

    if (a == NULL) 
      return NULL;

    if (type != ANY_ARC && a->GetType() != type) {
      srcarcs->Delete(P);
      return NULL;
    }
    return a;
}

GraphArc* GraphNode::NextArc( AvlPos*&P, ArcType type, ArcDirection &dir ) {
    GraphArc *a;

    // the iterator P does has no set identity -- it is switched
    // from the srcarcs to the dstarcs for dir ARC_INOUT
    for ( ; ; ) {
        a = srcarcs->Next(P);

        // end of the desired type sequence in the set
        if (a == NULL)
            break;
	if (type != ANY_ARC && a->GetType() != type) {
	  srcarcs->Delete(P);
	  break;
	}

        // if iteration over both src and dst arcs, then
        // the duplicate self arcs must be filtered out.
        if (dir != ARC_INOUT_IN || a->GetSrc() != this) return a;
    }

    if (dir == ARC_INOUT) {
        // for in/out direction, switch over to the dstarcs set.
        // retain direction to prevent duplication of self arcs.
        dir = ARC_INOUT_IN;
        return FirstArc(P,type, dir);
    }
    return NULL;
}

int GraphNode::ArcNum( ArcDirection dir ) {
    if (dir==ARC_OUT) return srcarcs->Size();
    if (dir==ARC_IN) return dstarcs->Size();
    return srcarcs->Size() + dstarcs->Size();
}

int GraphNode::ArcNum(ArcType type, ArcDirection d) {
    ArcDirection dir = d;
    if (type==ANY_ARC) return ArcNum(dir);

    GraphArc *a; AvlPos *P; int i=0;
    for ( a=FirstArc(P,type,dir); a; a=NextArc(P,type,dir))
        i++;
    return i;
}

char*	GraphNode::GetAttr( char *key ) {
    if (!attributes) return NULL;
    return attributes->GetAttr( key );
}

int	GraphNode::RemoveAttr( char *key ) {
    if (!attributes) return 0;
    char *attr = GetAttr( key );
    if (attr == NULL) return 0;
    attributes->RemoveAttr( key );
    return 1;
}

int	GraphNode::SetAttr( char *k, char *v ) {
    if (!attributes) attributes = new AttributeSet();
    return attributes->SetAttr( k,v );
}

char*	GraphNode::FirstAttr(AvlPos*&P) {
    if (!attributes) return NULL;
    return attributes->First(P);
}

char*	GraphNode::NextAttr(AvlPos*&P) {
    if (!attributes) return NULL;
    return attributes->Next(P);
}

int	GraphNode::NumAttrs() {
    if (!attributes) return 0;
    return attributes->Size();
}

int	GraphNode::PrintAttrs() {
    if (!attributes) return -1;
    attributes->Print();
    return 0;
}


// Perform a depth first traversal of the hierarchy, starting at this node. We
// follow arcs only if they are of the type given. At each node
// we visit, we call the function nfp() with argument nvp. At each arc we
// use for traversal, we call the function afp() with argument avp. Also, at
// each node we visit, the function checkfunc is called. If this function
// returns TRUE, the node is used in the traversal for further processing. If
// it returns FALSE, the node is not processed in the traversal, and that
// branch is pruned at that point.

void GraphNode::TreeDFS( ArcFlags& types, ArcDirection d,
                    VarArgFunRef nfp, void* nvp,
                    VarArgFunRef afp, void* avp,
                    IntFunction checkfunc, void* cvp ) {
    ArcDirection dir = d;
    int ntypes = types.Size();
    Stack s; /* = Stack(); */

    if (checkfunc == NULL)
        s.Push(this);
    else if (cvp == NULL && (*checkfunc)(this))
        s.Push(this);
    else if (cvp != NULL && (*checkfunc)( this, cvp ))
        s.Push(this);

    while (!s.IsEmpty()) {
        GraphNode* np = (GraphNode*)s.Pop();
        if (nfp != NULL){
            if (nvp == NULL) (*nfp)(np); else (*nfp)(np, nvp);
        }

        for ( int type=0; type<ntypes; type++ ) {
            if ( !types.GetBit(type) ) continue;

            AvlPos *P;
            GraphArc* ap = np->FirstArc( P, type, dir );
            while (ap != NULL)
            {
                if (afp != NULL)
                {
                    if (avp == NULL) (*afp)(ap); else (*afp)(ap, avp);
                }
                GraphNode* othernode = ap->GetOtherNode(np);
                if (othernode)
                {
                    if (checkfunc == NULL)
                        s.Push(othernode);
                    else if (cvp == NULL && (*checkfunc)(othernode))
                        s.Push(othernode);
                    else if (cvp != NULL && (*checkfunc)( othernode, cvp ))
                        s.Push(othernode);
                }
                ap = np->NextArc( P,type,dir );
            }
        }
    }
}

void GraphNode::TreeDFS( AdjList adj, ArcType type,
                    VarArgFunRef nfp, void* nvp,
                    VarArgFunRef afp, void* avp,
                    IntFunction checkfunc, void* cvp )
{
    Stack s; /* = Stack(); */
    if (checkfunc == NULL)
        s.Push(this);
    else if (cvp == NULL && (*checkfunc)(this))
        s.Push(this);
    else if (cvp != NULL && (*checkfunc)( this, cvp ))
        s.Push(this);

    while (!s.IsEmpty())
    {
        GraphNode* np = (GraphNode*)s.Pop();
        if (nfp != NULL)
        {
            if (nvp == NULL) (*nfp)(np); else (*nfp)(np, nvp);
        }
        AvlPos *P;
        GraphArc* ap = np->FirstAdj( P, adj );
        while (ap != NULL)
        {
            if ( (adj!=inArcs&&adj!=outArcs)
            || ANY_ARC==type || ap->GetType()==type )
            {
                if (afp != NULL)
                {
                    if (avp == NULL) (*afp)(ap); else (*afp)(ap, avp);
                }
                GraphNode* othernode = ap->GetOtherNode(np);
                if (othernode)
                {
                    if (checkfunc == NULL)
                        s.Push(othernode);
                    else if (cvp == NULL && (*checkfunc)(othernode))
                        s.Push(othernode);
                    else if (cvp != NULL && (*checkfunc)( othernode, cvp ))
                        s.Push(othernode);
                }
            }
            ap = np->NextAdj( P, adj );
        }
    }
}

// The same depth first traversal, except we do not have a list specified
// and we do have an arc direction. Note that it is illegal to do a tree
// traversal with a direction of both IN and OUT arcs

void GraphNode::TreeDFS( ArcType type, ArcDirection dir,
                    VarArgFunRef nfp, void* nvp,
                    VarArgFunRef afp, void* avp,
                    IntFunction checkfunc, void* cvp )
{
    AdjList list;
    if ( dir == ARC_IN )
        list = (type == editor->LevelArc()) ? inLevArcs : inArcs;
    else if ( dir == ARC_OUT )
        list = (type == editor->LevelArc()) ? outLevArcs : outArcs;
    else {
        mayday("GraphNode::TreeDFS - DFS illegal on IN & OUT arcs simultaneously");
        return;
    }
    TreeDFS( list, type, nfp, nvp, afp, avp, checkfunc, cvp );
}


// The same depth first traversal, except we do not have a type specified
// and we do not have a function to prune the branches of the traversal.

void GraphNode::TreeDFS( AdjList adj,
                    VarArgFunRef nfp, void* nvp,
                    VarArgFunRef afp, void* avp) {

// Generic depth first search on incoming arcs
Stack s; /* = Stack(); */
s.Push(this);
while (!s.IsEmpty())
{
    GraphNode* np = (GraphNode*)s.Pop();
    if (nfp != NULL)
    {
    if (nvp == NULL) (*nfp)(np); else (*nfp)(np, nvp);
    }
    AvlPos *P;
    GraphArc* ap = np->FirstAdj( P, adj );
    while (ap != NULL)
    {
        if (afp != NULL)
        {
            if (avp == NULL) (*afp)(ap); else (*afp)(ap, avp);
        }
        GraphNode* othernode = ap->GetOtherNode(np);
        if (othernode != NULL) s.Push(othernode);
        ap = np->NextAdj( P, adj );
    }
}
}

// Perform a breadth first traversal of the hierarchy, starting at this node.
// We follow nodes only if they are of the type given. At each node
// we visit, we call the function nfp() with argument nvp. At each arc we
// use for traversal, we call the function afp() with argument avp. Also, at
// each node we visit, the function checkfunc (if defined) is called. If this 
// function returns TRUE, the node is used in the traversal for further
// processing. If it returns FALSE, the node is not processed in the traversal,
// and that branch is pruned at that point. If the function is not defined,
// then all nodes are queued for further processing.

void GraphNode::TreeBFS( ArcFlags& types, ArcDirection d,
                    VarArgFunRef nfp, void* nvp,
                    VarArgFunRef afp, void* avp,
                    IntFunction checkfunc, void* cvp)
{
    ArcDirection dir = d;

    // Generic breadth first search on given arc type
    int ntypes = types.Size();

    Queue q; /* = Queue(); */
    if (checkfunc == NULL) q.Enqueue( this );
    else if (cvp == NULL && (*checkfunc)(this)) q.Enqueue(this);
    else if (cvp != NULL && (*checkfunc)( this, cvp )) q.Enqueue(this);
    while (!q.IsEmpty())
    {
        GraphNode* np = (GraphNode*)q.Dequeue();
        if (nfp != NULL)
        {
            if (nvp == NULL) (*nfp)(np); else (*nfp)(np, nvp);
        }

        for ( int type=0; type<ntypes; type ++ )
        {
            if ( !types.GetBit(type) ) continue;

            AvlPos *P;
            GraphArc* ap = np->FirstArc( P, type, dir );
            while (ap != NULL)
            {
                if (afp != NULL)
                {
                    if (avp == NULL) (*afp)(ap); else (*afp)(ap, avp);
                }
                GraphNode* othernode = ap->GetOtherNode(np);
                if ( othernode )
                {
                    if (checkfunc == NULL)
                        q.Enqueue(othernode);
                    else if (cvp == NULL && (*checkfunc)( othernode ))
                        q.Enqueue(othernode);
                    else if (cvp != NULL && (*checkfunc)( othernode, cvp ))
                        q.Enqueue(othernode);
                }
                ap = np->NextArc( P, type, dir );
            }
        }
    }
        if (q.Length()!=0) mayday("Bad queue! Leftover length is %d",q.Length());
}

void GraphNode::TreeBFS( AdjList adj, ArcType type,
                    VarArgFunRef nfp, void* nvp,
                    VarArgFunRef afp, void* avp,
                    IntFunction checkfunc, void* cvp)
{

    // Generic breadth first search on given arc type
    Queue q; /* = Queue(); */
    if (checkfunc == NULL) q.Enqueue( this );
    else if (cvp == NULL && (*checkfunc)(this)) q.Enqueue(this);
    else if (cvp != NULL && (*checkfunc)( this, cvp )) q.Enqueue(this);
    while (!q.IsEmpty())
    {
        GraphNode* np = (GraphNode*)q.Dequeue();
        if (nfp != NULL)
        {
            if (nvp == NULL) (*nfp)(np); else (*nfp)(np, nvp);
        }

        AvlPos *P;
        GraphArc* ap = np->FirstAdj( P, adj );
        while (ap != NULL)
        {
            if ( ( adj!=inArcs && adj!=outArcs )
            || ap->GetType()==type || ANY_ARC==type )
            {
                if (afp != NULL)
                {
                    if (avp == NULL) (*afp)(ap); else (*afp)(ap, avp);
                }
                GraphNode* othernode = ap->GetOtherNode(np);
                if ( othernode )
                {
                    if (checkfunc == NULL)
                        q.Enqueue(othernode);
                    else if (cvp == NULL && (*checkfunc)( othernode ))
                        q.Enqueue(othernode);
                    else if (cvp != NULL && (*checkfunc)( othernode, cvp ))
                        q.Enqueue(othernode);
                }
            }
            ap = np->NextAdj( P, adj );
        }
    }
    if (q.Length()!=0) mayday("Bad queue! Leftover length is %d", q.Length());
}

// The same depth first traversal, except we do not have a list specified
// and we do have an arc direction. Note that it is illegal to do a tree
// traversal with a direction of both IN and OUT arcs

void GraphNode::TreeBFS( ArcType type, ArcDirection dir,
                    VarArgFunRef nfp, void* nvp,
                    VarArgFunRef afp, void* avp,
                    IntFunction checkfunc, void* cvp )
{
    AdjList list;
    if ( dir == ARC_IN )
        list = (type == editor->LevelArc()) ? inLevArcs : inArcs;
    else if ( dir == ARC_OUT )
        list = (type == editor->LevelArc()) ? outLevArcs : outArcs;
    else {
        mayday("GraphNode::TreeBFS - BFS illegal on IN & OUT arcs simultaneously");
        return;
    }
    TreeBFS( list, type, nfp, nvp, afp, avp, checkfunc, cvp );
}

// The same breadth first traversal, except we do not have a type specified
// and we do not have a function to prune the branches of the traversal.

void GraphNode::TreeBFS( AdjList adj,
                    VarArgFunRef nfp, void* nvp,
                    VarArgFunRef afp, void* avp)
{
// Generic breadth first search on incoming arcs
Queue q; /* = Queue(); */
q.Enqueue(this);
while (!q.IsEmpty())
{
    GraphNode* np = (GraphNode*)q.Dequeue();
    if (nfp != NULL)
    {
    if (nvp == NULL) (*nfp)(np); else (*nfp)(np, nvp);
    }
    AvlPos *P;
    GraphArc* ap = np->FirstAdj( P, adj );
    while (ap != NULL)
    {
        if (afp != NULL)
        {
            if (avp == NULL) (*afp)(ap); else (*afp)(ap, avp);
        }
        GraphNode* othernode = ap->GetOtherNode(np);
        if (othernode != NULL) q.Enqueue(othernode);
        ap = np->NextAdj( P, adj );
    }
}
}

// Copy the immediate instance variables of the node. No arc info
// or other list info is copied. Do not copy loaded flag.

void GraphNode::Copy( GraphNode* dst )
{
    dst->SetLabel( nLa );
    dst->SetType( GetType()  );
    dst->SetLevel( GetLevel() );
    dst->SetNDCPosition( pos );
    dst->SetActiveNode( active );
}

// ---------------------------------------------------------------------
// Access function that determines whether or not the label of this node
// fits the regular expression given.
// ---------------------------------------------------------------------

RigiBool GraphNode::GrepNodeLabel(char* regexp) { return(grep(nLa, regexp)); }

// ---------------------------------------------------------------------


RigiBool GraphNode::operator==( const GraphNode& n ) const
{
    return ((RigiBool)(GetId() == n.GetId()));
}    


int GraphNode::Printout( int ) const
{
    Mess("%d  ",GetId());
    return 10;
}

unsigned int GraphNode::Hash() const
{
    return (unsigned int) GetId();
}

RigiBool GraphNode::isEqual( void* a ) const
{
    if ( a == NIL ) return RigiFalse;
    return *this==*(GraphNode*)a;
}
