//
// cl_rigistate.h - Header file that defines the state of the Rigi editor. This
// class is a controller class that contains most of the global information
// that is needed to define the global state of the editor.
//

// $Author: jmcdanie $
// $Date: 1996-10-26 00:17:19 $
// $Source: /root/cvs/Rigi/rigiedit/cl_rigistate.h,v $
// $Locker:  $
// $State: Exp $

//------------------------------------------------------------------------

#ifndef RIGISTATE_H
#define RIGISTATE_H

// Include class definitions for various objects that the editor manipulates.


#ifndef OBJECTH
#include "adt/object.h"	// For RigiObject class
#endif
#ifndef _CCHEADER_H_
#include "CCheader.h"	// for RigiBool class
#endif
#ifndef GRAPHDEF_H
#include "cl_graphdef.h"
#endif
#ifndef ARCFLAGS_H
#include "cl_arcflags.h" // for ArcFlags class
#endif

class NodeNameSet;	// cl_graphnameset (for access-by-name)
class RigiNode;		// cl_riginode
class RigiNodeList;	// cl_riginode
class RigiWindow;	// cl_rigiwindow
class RigiWindowList;	// cl_rigiwindow
class RigiArc;		// cl_rigiarc

class ArcSet;		// cl_grapharcset
class NodeSet;		// cl_graphnodeset
class RigiNodeSet;	// cl_riginodeset
class AvlPos;

typedef enum {SCALE_NONE, SCALE_WINDOW, SCALE_BOX, SCALE_ONE, SCALE_FACTOR}
        RigiScale;
typedef enum {CON_NONE, CON_HORIZ, CON_VERT} RigiConstraint;
typedef enum {REPORT_SIMPLE, REPORT_VERBOSE} RigiReport;

class EditorState : public RigiObject
{
    private:

        // Selection and filtered sets/lists for nodes/arcs

        NodeSet		*selectnodes;	// set of selected nodes
        ArcSet		*selectarcs;	// set of selected arcs
        RigiArc		*selectarc;	// Selected arc
        RigiNodeList	*filterlist;	// List of filtered nodes
        RigiNodeSet	*filterset;	// Set of filtered nodes

        GraphNode	*lastselectednode;

        NodeSet		*globalnodeset; // set of nodes in the editor
        ArcSet		*globalarcset;	// set of arcs in the editor
        RigiWindowList	*globalwinlist;	// List of windows in the editor
        NodeNameSet	*globalnameset;	// set of node names in the editor

        // Parameters that determine how Rigi operates

        int            over_depth;      // Depth of overview windows
        int            proj_depth;      // Depth of projection windows
        int            hier_depth;      // Depth for dependency trees

        int            scale;           // Scale factor
        RigiScale      scale_function;  // Scale function

        RigiConstraint constraint;      // Constraint type

        RigiBool       display_anno;    // Display annotation???
        RigiBool       display_message; // Display messages
        RigiBool       display_command; // Display command area
        RigiBool       display_menu;    // Display the menu

        RigiBool       move_all;	// Multinode moves or single node

        int            grid_size;       // Grid size for gridding operations
        float          high_threshold;  // High threshold for operations
        float          low_threshold;   // Low threshold for operations
        RigiReport     report_type;     // Type of report generation
        int            arc_type;        // Arc type for operations
        ArcFlags       arc_set;         // Arc type set for operations
        int            level_arc;	// arctype interpreted as level
        int            composite_arc;	// arctype interpreted as composite
        int            collapse_node;	// nodetype used in collapse ops
        RigiBool       inherit_filters; // Inherit filters
        int            objId;       // last assigned node/arc id

    public:

        // Constructor and destructor

        EditorState();
        ~EditorState();

        // Initialize the dynamic stuff for the editor state
    
        void           Init();

        // Access functions for control parameters

        int            GetOverviewDepth() const;
        int            GetProjectionDepth() const;
        int            GetHierarchyDepth() const;

        RigiScale      GetScaleType() const;
        int            GetScaleFactor() const;
        RigiConstraint GetConstraintType() const;

        RigiBool       GetDisplayAnnotation() const;
        RigiBool       GetDisplayMessage() const;
        RigiBool       GetDisplayCommand() const;
        RigiBool       GetDisplayMenu() const;

        int            GetGridSize() const;
        float          GetHighThreshold() const;
        float          GetLowThreshold() const;
        RigiReport     GetReportGeneration() const;
        int            GetArcType() const;
        RigiBool       GetInheritFilters() const;
        RigiBool       GetMoveUnison() const;

        void           SetOverviewDepth(int);
        void           SetProjectionDepth(int);
        void           SetHierarchyDepth(int);

        void           SetScaleType(RigiScale);
        void           SetScaleFactor(int);
        void           SetConstraintType(RigiConstraint);

        void           SetDisplayAnnotation(RigiBool);
        void           SetDisplayMessage(RigiBool);
        void           SetDisplayCommand(RigiBool);
        void           SetDisplayMenu(RigiBool);

        void           SetGridSize(int);
        void           SetHighThreshold(float);
        void           SetLowThreshold(float);
        void           SetReportGeneration(RigiReport);
        void           SetArcType(int);
        void           SetInheritFilters(RigiBool);
        void           SetMoveUnison(RigiBool);

        // Methods for access to the arc type set.

        unsigned char InArcSet(int arctype) const;
        unsigned char InArcSet(char *arctypename) const;
        void          SetArcSet(int arctype, unsigned char value);
        void          SetArcSet(char *arctypename, unsigned char value);
        void          ClearArcSet();
        void          SetAllArcSet();

        // Methods for controlling "important" arc and node types

        int	        LevelArc()		{ return level_arc; }
        int	        CompositeArc()		{ return composite_arc; }
        int	        CollapseNode()		{ return collapse_node; }
        void	    SetLevelArc( int t )	{ level_arc = t; }
        void	    SetCompositeArc( int t)	{ composite_arc = t; }
        void	    SetCollapseNode( int t)	{ collapse_node = t; }


        // Methods for selection mask processing

        int         GetSelectionMask() const;
        void        SetSelectionMask( int mask );

        // Node and arc selection routines

        void        SelectNode( GraphNode* );
        void        SelectNode( RigiNode* );
        void        DeselectNode( GraphNode* );
        void        DeselectNode( RigiNode* );

        void        InvertSelectionNodes( RigiWindow*, Point, Point );
        void        DrawNodes( NodeSet*);
        void        MoveNodes( NodeSet*, int x, int y );

        void        SelectArc( RigiArc* );
        void        DeselectArc();
        RigiArc*    GetSelectedArc();

        void		SelectArc( GraphArc* );
        void		DeselectArc( GraphArc* );
        void		SelectArcs( RigiWindow*, Point, Point );
        GraphArc*	FirstSelectedArc();
        GraphArc*	NextSelectedArc();

        GraphNode*	FirstSelectedNode( AvlPos*& );
        GraphNode*	NextSelectedNode( AvlPos*& );
        GraphNode*	LastSelectedNode();
        int		    NumSelectedNodes();

        void        GetSelectionList( NodeList& );
        void		GetSelectionList( RigiNodeList&, RigiWindow* );
        void		EmptySelection();
        void		InvertSelection();

        // Node filtering routines

        RigiNode*   FirstFilteredNode();
        RigiNode*   NextFilteredNode();
        int         NumFilteredNodes();

        void        SetFilter(RigiNodeList&);
        void        SetFilter(RigiNode*);
        void 	    SetFilter(RigiNodeList&, RigiWindow*);
        void        ResetFilter();
        void        ResetFilter(RigiNode*);
        void	    ResetFilter(RigiWindow*);

        // Node and arc bookkeeping functions

        void        InitObjId()       {objId = 1<<20;}
        void        SetObjId(int id)  {objId = id;}
        int         GetObjId() const  {return objId; }
        int         NextObjId()       {return ++objId;}

        RigiBool    AddNode( GraphNode* );
        RigiBool    RemoveNode( GraphNode* );
        RigiBool    ChangeNodeName( GraphNode*,char* );
        int         EmptyNodes();

        GraphNode*  FirstNode( AvlPos*& );
        GraphNode*  NextNode( AvlPos*& );
        int         NumNodes();

        GraphNode*  FindNode( const NodeId );
        GraphNode*  FindNode( char* label  );
        int         FindNodes( char*, NodeList& );

        RigiBool    AddArc(GraphArc&);
        GraphArc*   RemoveArc(const ArcId);
        GraphArc*   FindArc(const ArcId);
        int         EmptyArcs();

        int	        PrintNodes();
        int         PrintNames( char *name = NULL );
        int         PrintArcs();

        // Window bookkeeping functions and list iterators

        void        AddWindow(RigiWindow*);
        RigiWindow* RemoveWindow(RigiWindow *);
        RigiWindow* FindWindow(RigiWindow *);
        RigiWindow* FirstWindow();
        RigiWindow* NextWindow();
        RigiWindow* GetWindowByNumber(int);

        // Display all the windows
        void RigiDisplay();

        // Close windows
        void        CloseWindow(RigiWindow *);
        void        CloseAll();
        void        CloseBaseWindow();

        // Wipe out the entire graph (except root node)
        void        Reset();

        // State archiving access functions. NOTE: none of the node, arc, and
        // window lists are including in the archive. Only the selection and
        // filtering information is archived.

        int         WriteArchive(FILE* fp);
        int         ReadArchive(FILE* fp);
};

#endif

